pre code.hljs {
  display: block;
  overflow-x: auto;
  padding: 1em;
}

code.hljs {
  padding: 3px 5px;
}

.hljs {
  color: #e0e2e4;
}

.hljs-keyword,
.hljs-literal,
.hljs-selector-id,
.hljs-selector-tag {
  color: #93c763;
}

.hljs-number {
  color: #ffcd22;
}

.hljs-attribute {
  color: #668bb0;
}

.hljs-link,
.hljs-regexp {
  color: #d39745;
}

.hljs-meta {
  color: #557182;
}

.hljs-addition,
.hljs-built_in,
.hljs-bullet,
.hljs-emphasis,
.hljs-name,
.hljs-selector-attr,
.hljs-selector-pseudo,
.hljs-subst,
.hljs-tag,
.hljs-template-tag,
.hljs-template-variable,
.hljs-type,
.hljs-variable {
  color: #8cbbad;
}

.hljs-string,
.hljs-symbol {
  color: #ec7600;
}

.hljs-comment,
.hljs-deletion,
.hljs-quote {
  color: #818e96;
}

.hljs-selector-class {
  color: #a082bd;
}

.hljs-doctag,
.hljs-keyword,
.hljs-literal,
.hljs-name,
.hljs-section,
.hljs-selector-tag,
.hljs-strong,
.hljs-title,
.hljs-type {
  font-weight: 700;
}

.hljs-class .hljs-title,
.hljs-code,
.hljs-section,
.hljs-title.class_ {
  color: #fff;
}
